#!/usr/bin/env python

import re
import sets

import pprint as pp

def main(base, taxonomy_file):
    seqinfo = base+"_seqinfo.csv"
    output_tax = base+"_taxonomy.csv"
    
    Getaxnseq().gg_taxonomy_builder(taxonomy_file, output_tax, seqinfo)
    
    return seqinfo, output_tax

class Getaxnseq:

    def __init__(self): pass
    
    def taxonomy_line(self, level_index, taxon_array):
        if level_index == 0:
            return '%s,Root,kingdom,%s,%s,%s,,,,,,' % (taxon_array[level_index], taxon_array[level_index], 'Root', taxon_array[level_index])
        elif level_index == 1:
            return '%s,%s,phylum,%s,%s,%s,%s,,,,,' % (taxon_array[level_index], taxon_array[level_index-1], taxon_array[level_index], 'Root', taxon_array[0], taxon_array[1])
        elif level_index == 2:
            return '%s,%s,class,%s,%s,%s,%s,%s,,,,' % (taxon_array[level_index], taxon_array[level_index-1], taxon_array[level_index], 'Root', taxon_array[0], taxon_array[1], taxon_array[2])
        elif level_index == 3:
            return '%s,%s,order,%s,%s,%s,%s,%s,%s,,,' % (taxon_array[level_index], taxon_array[level_index-1], taxon_array[level_index], 'Root', taxon_array[0], taxon_array[1], taxon_array[2], taxon_array[3])
        elif level_index == 4:
            return '%s,%s,family,%s,%s,%s,%s,%s,%s,%s,,' % (taxon_array[level_index], taxon_array[level_index-1], taxon_array[level_index], 'Root', taxon_array[0], taxon_array[1], taxon_array[2], taxon_array[3], taxon_array[4])
        elif level_index == 5:
            return '%s,%s,genus,%s,%s,%s,%s,%s,%s,%s,%s,' % (taxon_array[level_index], taxon_array[level_index-1], taxon_array[level_index], 'Root', taxon_array[0], taxon_array[1], taxon_array[2], taxon_array[3], taxon_array[4], taxon_array[5])
        elif level_index == 6:
            return '%s,%s,species,%s,%s,%s,%s,%s,%s,%s,%s,%s' % (taxon_array[level_index], taxon_array[level_index-1], taxon_array[level_index], 'Root', taxon_array[0], taxon_array[1], taxon_array[2], taxon_array[3], taxon_array[4], taxon_array[5], taxon_array[6])
        else:
            raise Exception("Programming error, found too many levels!")


    def gg_taxonomy_builder(self, taxonomy_file, output_taxonomy, output_seqinfo):
        first_pass_id_and_taxonomies = []
        meaningless_taxonomic_names = sets.Set(['k__', 'd__', 'p__', 'c__', 'o__', 'f__', 'g__', 's__'])

        for entry in open(taxonomy_file):
            # split the entire line including the  on '; '
            split = entry.rstrip().split('; ')

            # split out the taxon ID from the first split of above
            taxon_id = split[0].split()[0]
            first_taxon = split[0].split()[1]
            tax_split = [first_taxon] + split[1:]

            # Replace spaces with underscores e.g. 'Candidatus my_genus'
            for idx, item in enumerate(tax_split):
                tax_split[idx] = re.sub('\s+', '_', item)

            # Remove 'empty' taxononomies e.g. 's__'
            tax_split = [item for item in tax_split if item not in meaningless_taxonomic_names]

            # Add this fixed up list to the list
            first_pass_id_and_taxonomies.append([taxon_id]+tax_split)

        # Find taxons that have multiple parents, building a hash of parents as we go (i.e. a tree of taxonomies embedded in a hash)
        #
        # Assumes that no two taxonomic labels are the same when they are from different
        # taxonomic levels. When there are children with multiple parents at the
        # same taxonomic label then these are warned about and worked around.
        parents = {} #hash of taxon to immediate parent
        known_duplicates = sets.Set([])
        for j, array in enumerate(first_pass_id_and_taxonomies):
            taxonomy = array[1:]
            for i, tax in enumerate(taxonomy):
                if i==0: continue #top levels don't have parents
                ancestry = taxonomy[i-1]
                if parents.has_key(tax):
                    if parents[tax] != ancestry:

                        dup = "%s%s" %(parents[tax], ancestry)
                        # don't report the same problem several times
                        if dup not in known_duplicates:
                            print " %s '%s' with multiple parents %s and %s" % (array[0], tax, parents[tax], ancestry)
                            known_duplicates.add(dup)

                        # fix the current one
                        new_name_id = 1
                        new_name = "%se%s" % (tax, new_name_id)

                        while parents.has_key(new_name) and parents[new_name] != ancestry:
                            new_name_id += 1
                            new_name = "%se%s" % (tax, new_name_id)

                        first_pass_id_and_taxonomies[j][i+1] = new_name
                        taxonomy[i] = new_name
                        parents[new_name] = ancestry
                else:
                    # normal case, seeing a new taxon and parent for the first time

                    parents[tax] = ancestry

        #pp.pprint(parents)
        #pp.pprint(first_pass_id_and_taxonomies)

        # Write the sequence file
        with open(output_seqinfo, 'w') as seqout:
            # write header
            seqout.write('seqname,tax_id\n')
            # write each taxonomic association
            for array in first_pass_id_and_taxonomies:
                seqout.write("%s,%s\n" % (array[0], array[-1]))

        # Write the taxonomy file
        noted_taxonomies = sets.Set([])
        with open(output_taxonomy, 'w') as seqout:
            # write header and root line
            seqout.write('tax_id,parent_id,rank,tax_name,root,kingdom,phylum,class,order,family,genus,species\n')
            seqout.write('Root,Root,root,Root,Root,,,,,,,\n')
            # write all the taxonomies
            for array in first_pass_id_and_taxonomies:
                taxons = array[1:]



                for i, tax in enumerate(taxons):
                    line = self.taxonomy_line(i, taxons[:(i+1)])

                    if line not in noted_taxonomies:
                        seqout.write(line+"\n")
                        noted_taxonomies.add(line)





import subprocess
import os
import json
import timeit
import logging

from Bio import SeqIO

from graftm.classify import Classify
from graftm.assembler import TaxoGroup
from graftm.housekeeping import HouseKeeping


class Pplacer:
    ### Contains function related to processing alignment files to jplace files
    ### and running comparisons between forward and revere reads if reverse
    ### reads are provided.

    def __init__(self, refpkg):
        self.refpkg = refpkg
        self.hk = HouseKeeping()

    # Run pplacer
    def pplacer(self, output_file, output_path, input_path, threads):
        ## Runs pplacer on concatenated alignment file
        cmd = "pplacer -j %s --verbosity 0 --out-dir %s -c %s %s" % (str(threads), output_path, self.refpkg, input_path) # Set command
        logging.debug("Running command: %s" % cmd)
         # Log it
        subprocess.check_call(cmd, shell=True) # Run it
        output_path = '.'.join(input_path.split('.')[:-1]) + '.jplace'
        return output_path

    def alignment_merger(self, alignment_files, output_alignment_path):
        ## Concatenate aligned read_files into one file. Each read with it's
        ## own unique identifier assigning it to a particular origin file
        alias_hash = {} # Set up a hash with file names and their unique identifier
        file_number = 0 # file counter (unique identifier)
        with open(output_alignment_path, 'w') as output:
            for alignment_file in alignment_files: # For each alignment
                alignments = list(SeqIO.parse(open(alignment_file, 'r'), 'fasta')) # read list
                for record in alignments: # For each record in the read list
                    record.id = record.id + '_' + str(file_number) # append the unique identifier to the record id
                SeqIO.write(alignments, output, "fasta") # And write the reads to the file
                alias_hash[str(file_number)] = {'output_path': os.path.join(os.path.dirname(alignment_file),'placements.jplace') ,
                                             'place': []}
                file_number += 1
        return alias_hash

    def jplace_split(self, jplace_file, alias_hash, summary_dict):
        ## Split the jplace file into their respective directories

        # Load the placement file
        placement_file = json.load(open(jplace_file))

        # Parse the placements based on unique identifies appended to the end
        # of each read
        for placement in placement_file['placements']: # for each placement
            hash = {} # create an empty hash
            for alias in alias_hash: # For each alias, append to the 'place' list each read that identifier
                hash = {'p': placement['p'],
                        'nm': [nm for nm in placement['nm'] if nm[0].split('_')[-1] == alias]}
                alias_hash[alias]['place'].append(hash)

        # Write the jplace file to their respective file paths.
        jplace_path_list = []
        for alias in alias_hash:
            output = {'fields': placement_file['fields'],
                      'version': placement_file['version'],
                      'tree':  placement_file['tree'],
                      'placements': alias_hash[alias]['place'],
                      'metadata': placement_file['metadata']}
            with open(alias_hash[alias]['output_path'], 'w') as output_path:
                json.dump(output, output_path, ensure_ascii=False)
            jplace_path_list.append(alias_hash[alias]['output_path'])
        summary_dict['jplace_path_list'] = jplace_path_list
        return summary_dict

    def place(self, summary_dict, files, args):
        ## Pipeline taking multiple alignment files and returning multiple
        ## placement and guppy files, as well as the comparison between forward
        ## and reverse reads, if the reverse pipeline is selected

        start = timeit.default_timer() # Start placement timer

        # Merge the alignments so they can all be placed at once.
        alias_hash = self.alignment_merger(summary_dict['seqs_list'], files.comb_aln_fa())
        
        # Run pplacer on merged file
        jplace = self.pplacer(files.jplace_output_path(), args.output_directory, files.comb_aln_fa(), args.threads)
        logging.info("Placements finished")

        stop = timeit.default_timer() # stop placement timer and log
        summary_dict['place_t'] = str( int(round((stop - start), 0)) )

        # Split the jplace file
        summary_dict = self.jplace_split(jplace, alias_hash, summary_dict)
        
        #Read the json of refpkg
        logging.info("Reading classifications")
        tax_descr=json.load(open(self.refpkg+'/CONTENTS.json'))['files']['taxonomy']
        classifications=Classify(os.path.join(self.refpkg,tax_descr)).assignPlacement(jplace, args.placements_cutoff, 'reads', summary_dict['resolve_placements'])
        self.hk.delete([jplace])# Remove combined split, not really useful
        logging.info("Reads classified.")
        
        # If the reverse pipe has been specified, run the comparisons between the two pipelines. If not then just return.

        for idx, base in enumerate(summary_dict['base_list']):
            if summary_dict['reverse_pipe']:
                summary_dict[base] = Compare().compare_hits(summary_dict[base], base)

                forward_gup=classifications.pop(sorted(classifications.keys())[0]) 
                reverse_gup=classifications.pop(sorted(classifications.keys())[0])
                summary_dict[base] = Compare().compare_placements(forward_gup,
                                                                  reverse_gup,
                                                                  summary_dict[base],
                                                                  args.placements_cutoff)

            elif not summary_dict['reverse_pipe']: # Set the trusted placements as
                summary_dict[base]['trusted_placements'] = {}

                for read, entry in classifications[str(idx)].iteritems():
                    summary_dict[base]['trusted_placements'][read] = entry['placement']
        return summary_dict

class Compare:
    ### Functions for comparing forward and reverse read hits and placements

    def __init__(self): pass

    def compare_hits(self, hash, file_name):
        ## Take a paired read run, and compare hits between the two, report the
        ## number of hits each, the crossover, and a
        def check_reads(read_lists):
            for read_list in read_lists:
                read_check = [read for read in read_list if read.startswith('FCC')]
                if read_check:
                    return True
                elif not read_check:
                    continue
            return False
        # Read in the read names for each file
        check = check_reads([hash['forward']['reads'].keys(), hash['reverse']['reads'].keys()])
        if check:
            forward_read_names = [x.replace('/1', '') for x in hash['forward']['reads'].keys()]
            reverse_read_names = [x.replace('/2', '') for x in hash['reverse']['reads'].keys()]
        elif not check:
            forward_read_names = set(hash['forward']['reads'].keys())
            reverse_read_names = set(hash['reverse']['reads'].keys())
        # Report and record the crossover
        crossover_hits = [x for x in forward_read_names if x in reverse_read_names]

        hash['crossover'] = crossover_hits
        # Check if there are reads to continue with.
        if len(crossover_hits) > 0:
            logging.info("%s reads found that crossover in %s" % (str(len(crossover_hits)),
                                                                       file_name))
        elif len(crossover_hits) == 0:
            logging.info("%s reads found that crossover in %s, No reads to use!" % (str(len(crossover_hits)),
                                                                                           file_name))            
        else:
            raise Exception
        # Return the hash

        return hash 

    def compare_placements(self, forward_gup, reverse_gup, hash, placement_cutoff):
        ## Take guppy placement file for the forward and reverse reads, compare
        ## the placement, and make a call as to which is to be trusted. Return
        ## a list of trusted reads for use by the summary step in GraftM
        # Report and record trusted placements
        comparison_hash = {'trusted_placements': {}} # Set up a hash that will record info on each placement
        for read in hash['crossover']:
            if read.startswith('FCC'):
                f_read = read + '/1'
                r_read = read + '/2'
            else:
                f_read = read
                r_read = read
            # Check read was placed
            if forward_gup.get(f_read) is None or reverse_gup.get(r_read) is None:
                logging.info('Warning: %s was not inserted into tree' % str(f_read))
                continue # Skip read for now
            comparison_hash[read] = {} # make an entry for each read
            comparison_hash['trusted_placements'][read] = [] # Set up a taxonomy entry in trusted placements

            if len(forward_gup[f_read]['placement']) == len(reverse_gup[r_read]['placement']): # If the level of placement matches
                comparison_hash[read]['rank_length_match'] = True # Store True
            elif len(forward_gup[f_read]['placement']) != len(reverse_gup[r_read]['placement']):
                comparison_hash[read]['rank_length_match'] = False # Otherwise store False
            else:
                raise Exception('Programming Error: Comparison of placement resolution')
            for idx, (f_rank, r_rank) in enumerate(zip(forward_gup[f_read]['placement'], reverse_gup[r_read]['placement'])): # For the each rank in the read placement                    
                if f_rank == r_rank: # If the classification at this rank matches
                    comparison_hash[read]['all_ranks_match'] = True # Maintain the all ranks match are true
                    if comparison_hash[read]['rank_length_match']: # If both reads are to the same resolution
                        comparison_hash['trusted_placements'][read].append(f_rank) # Just append that rank
                    elif not comparison_hash[read]['rank_length_match']: # But if they are not at the same resolution
                        # check if we've reached the end of the forward or reverse taxonomy list
                        # and if we have, append the tail end of one with the longer taxonomy, because
                        # we can trust these placements are past the overall threshold,
                        # but one has higher resolution and so is more useful to us.
                        if len(forward_gup[f_read]['placement'][idx:]) == 1:
                            comparison_hash['trusted_placements'][read] += reverse_gup[r_read]['placement'][idx:]
                        elif len(reverse_gup[r_read]['placement'][idx:]) == 1:
                            comparison_hash['trusted_placements'][read] += forward_gup[f_read]['placement'][idx:]
                        elif len(forward_gup[f_read]['placement'][idx:]) > 1 and len(reverse_gup[r_read]['placement'][idx:]) > 1:
                            comparison_hash['trusted_placements'][read].append(f_rank)
                        else:
                            raise Exception('Programming Error')
                elif f_rank != r_rank: # Otherwise if the classification doesn't match
                    comparison_hash[read]['all_ranks_match'] = False # Store that all ranks do not match up
                    forward_confidence = forward_gup[f_read]['confidence'][idx] # Store confidence values for both directions
                    reverse_confidence = reverse_gup[r_read]['confidence'][idx]
                    if float(forward_confidence) > float(reverse_confidence): # If the forward read has more confidence
                        comparison_hash['trusted_placements'][read] += forward_gup[f_read]['placement'][idx:] # Store the taxonomy of that read from that point on
                        break
                    elif float(reverse_confidence) > float(forward_confidence): # Do the opposite if reverse read has more confidence
                        comparison_hash['trusted_placements'][read] += reverse_gup[r_read]['placement'][idx:]
                        break
                    elif float(reverse_confidence) == float(forward_confidence): # If they are of the same value
                        break # Do nothing, because a decision cannot be made if the confidence is equal.
                    else:
                        raise Exception('Programming Error: Comparing confidence values')
                else:
                    raise Exception('Programming Error: Comparison of placement resolution')               

        hash['comparison_hash'] = comparison_hash
        return hash # Return the hash




class Stats_And_Summary:

    def __init__(self): pass

    def coverage_of_hmm(self, hmm, count_table, coverage_table, avg_read_length):

        for line in open(hmm):

            if line.startswith('LENG'):
                length = line.split()[1]

        with open(coverage_table, 'w') as ct:
            write = ['#ID     20110816_S1D    ConsensusLineage\n']
            for line in open(count_table):

                    if line.startswith('#'):
                        continue

                    splt = line.split()

                    cov = str(round((float(splt[1])*float(avg_read_length)) / float(length), 3))

                    write.append("%s\t%s\t%s\n" % (
                                    splt[0],
                                    cov,
                                    splt[2]))
            for entry in write:
                ct.write(entry)

    def readTax(self, placements, output_path):
        with open(output_path, 'w') as out:
            for read, tax in placements.iteritems():
                out.write("%s\t%s\n" % (read, '; '.join(tax)))

    def build_basic_statistics(self, summary_hash, output, pipe):
        
        def compile_run_stats(hash):
            files = []
            tot_16S = []
            tot_18S = []
            cont_18S = []
            search_step = []
            placed_reads = 0

            aln_step = []
            euk_check_step = []
            for base in hash['base_list']:
                if hash['reverse_pipe']:
                    files += [base + ' forward', base + ' reverse']
                    tot_16S += [str(len(hash[base]['forward']['reads'].keys())), str(len(hash[base]['reverse']['reads'].keys()))]
                    try:
                        tot_18S += [str(hash[base]['forward']['euk_uniq'] + hash[base]['forward']['euk_contamination']), 
                                    str(hash[base]['reverse']['euk_uniq'] + hash[base]['reverse']['euk_contamination'])]
                    except:
                        tot_18S += ['N/A', 'N/A']
                    try:
                        cont_18S += [str(summary_hash[base]['forward']['euk_contamination']), str(summary_hash[base]['reverse']['euk_contamination'])] 
                    except:
                        cont_18S = ['N/A']
                    search_step += [hash[base]['forward']['search_t'], hash[base]['reverse']['search_t']]
                    aln_step += [hash[base]['forward']['aln_t'], hash[base]['reverse']['aln_t']]
                    euk_check_step += [hash[base]['forward']['euk_check_t'], hash[base]['reverse']['euk_check_t']]
                elif not hash['reverse_pipe']:
                    files += [base]
                    tot_16S += [str(len(hash[base]['reads'].keys()))]
                    try:
                        placed_reads+=int(tot_16S[0])-int(hash[base]['euk_contamination'])
                    except:

                        placed_reads+=int(tot_16S[0])
                    try:
                        tot_18S += [str(hash[base]['euk_contamination'] + hash[base]['euk_uniq'])]

                    except:
                        tot_18S += ['N/A']
                    try:
                        cont_18S += [str(summary_hash[base]['euk_contamination'])]
                    except:
                        cont_18S += ['N/A']
                    search_step += [hash[base]['search_t']]
                    aln_step += [hash[base]['aln_t']]
                    euk_check_step += [hash[base]['euk_check_t']]
                else:
                    raise Exception('Programming Error')
            return '\t'.join(files), '\t'.join(tot_16S), '\t'.join(tot_18S), '\t'.join(cont_18S), str(placed_reads), '\t'.join(files), '\t'.join(search_step), '\t'.join(aln_step), '\t'.join(euk_check_step), hash['place_t'], hash['summary_t'], hash['all_t']
        

        stats = """Basic run statistics (count):

                             Files:\t%s
Total number of 16S reads detected:\t%s
Total number of 18S reads detected:\t%s
    'Contaminant' eukaryotic reads:\t%s
              reads placed in tree:\t%s


Runtime (seconds):
                             Files:\t%s
                       Search step:\t%s
                    Alignment step:\t%s
              Eukaryote check step:\t%s

               Tree insertion step:\t%s
                 Summarising steps:\t%s
                     Total runtime:\t%s

    """ % compile_run_stats(summary_hash)
        
        with open(output, 'w') as stats_file:
            stats_file.write(stats)


    def otu_builder(self, hash_list, output_list, base_list, combined_output):
        ## A function that takes a hash of trusted placements, and compiles them
        ## into an OTU-esque table.
        # Start an output list and hash that will contain consolidated numbers
        # for each taxonomic rank...    
        all_hashes={}
        ps=[]
        for idx, base in enumerate(base_list):
            output_table = [['#ID', base, 'ConsensusLineage']]
            c_hash       = {}
            rank_id      = 0
            # Consolidate the placement list into a simplified hash with taxonomic
            # rank, and number. Simple.
            p=['; '.join(x) for x in hash_list[idx].values()]
            ps.append(p)
            for placement_list in p:
                try:
                    c_hash[placement_list] += 1
                except:
                    c_hash[placement_list] = 1
                    all_hashes[placement_list]=[]
            # Make an entry into the output list
            for rank, count in c_hash.iteritems():
                output_table.append([str(rank_id), str(count), rank])
                rank_id += 1
            # And write to a file
            with open(output_list[idx], 'w') as out:
                for line in output_table:
                    out.write('\t'.join(line)+'\n')
        output_table = [['#ID', '\t'.join(base_list), 'ConsensusLineage']]
        rank_id=0
        for p in ps:
            for place in all_hashes.keys():
                try:
                    all_hashes[place].append(p.count(place))
                except:
                    all_hashes[place].append(0)
        for key, item in all_hashes.iteritems():
            output_table.append([str(rank_id), '\t'.join([str(x) for x in item]), key])
            rank_id+=1
        with open(combined_output, 'w') as out:
            for line in output_table:
                out.write('\t'.join(line)+'\n')

            
        
        


#!/usr/bin/env python

import timeit
import os
import logging

from graftm.graftm_output_paths import GraftMFiles
from graftm.extract_sequences import Extract
from graftm.hmmer import Hmmer
from graftm.housekeeping import HouseKeeping
from graftm.summarise import Stats_And_Summary
from graftm.pplacer import Pplacer
from graftm.krona_from_community_profiles import KronaBuilder
from graftm.assembler import TaxoGroup
from graftm.create import Create

class Run:
    ### Functions that make up pipelines in GraftM

    def __init__(self, args):
        self.args = args
        self.setattributes(self.args)

    def setattributes(self, args):
        self.kb = KronaBuilder()
        self.hk = HouseKeeping()
        self.s = Stats_And_Summary()
        self.tg = TaxoGroup()
        self.e = Extract()
        if args.subparser_name == 'graft':
            self.hk.set_attributes(self.args)
            self.h = Hmmer(self.args.search_hmm_files, self.args.aln_hmm_file)
            self.sequence_pair_list, self.input_file_format = self.hk.parameter_checks(args)
            if hasattr(args, 'reference_package'):
                self.p = Pplacer(self.args.reference_package)

    def protein_pipeline(self, base, summary_dict, sequence_file, direction):
        'The main pipeline for GraftM finding protein hits in the input sequence'
        # Set a variable to store the run statistics, to be added later to
        # the summary_dict
        if direction:
            run_stats = summary_dict[base][direction]
        elif not direction:
            run_stats = summary_dict[base]
        else:
            raise Exception('Programming Error: Assigning run_stats hash')
        # Tell user what is being searched with what
        logging.info('Searching %s' % (os.path.basename(sequence_file)))
        # Search for reads using hmmsearch
        hit_reads, run_stats = self.h.p_search(self.gmf,
                                               self.args,
                                               run_stats,
                                               base,
                                               self.input_file_format,
                                               sequence_file)
        if not hit_reads:
            return summary_dict, False
        # Align the reads.
        logging.info('Aligning reads to reference package database')
        hit_aligned_reads, run_stats = self.h.align(self.gmf,
                                                    self.args,
                                                    run_stats,
                                                    base,
                                                    hit_reads)
        # Set these paramaters as N/A 'cos they don't apply to the protein pipeline
        run_stats['n_contamin_euks'] = 'N/A'
        run_stats['n_uniq_euks'] = 'N/A'
        run_stats['euk_check_t'] = 'N/A'
        if direction:
            summary_dict[base][direction] = run_stats
        elif not direction:
            summary_dict[base] = run_stats
        else:
            raise Exception('Programming Error: Logging %s hash' % direction)

        return summary_dict, hit_aligned_reads

    def dna_pipeline(self, base, summary_dict, sequence_file, direction):
        'The main pipeline for GraftM searching for DNA sequence'
        # Set a variable to store the run statistics, to be added later to
        # the summary_dict
        if direction:
            run_stats = summary_dict[base][direction]
        elif not direction:
            run_stats = summary_dict[base]
        else:
            raise Exception('Programming Error: Assigning run_stats hash')

        # Search for reads using nhmmer
        logging.info('Searching %s' % os.path.basename(sequence_file))
        hit_reads, run_stats = self.h.d_search(self.gmf,
                                               self.args,
                                               run_stats,
                                               base,
                                               self.input_file_format,
                                               sequence_file,
                                               summary_dict['euks_checked'] )
        
        if not hit_reads:
            return summary_dict, False
        
        # Otherwise, run through the alignment
        logging.info('Aligning reads to reference package database')
        hit_aligned_reads, run_stats = self.h.align(self.gmf,
                                                    self.args,
                                                    run_stats,
                                                    base,
                                                    hit_reads)
        if direction:
            summary_dict[base][direction] = run_stats
        elif not direction:
            summary_dict[base] = run_stats
        else:
            raise Exception('Programming Error: Logging %s hash' % direction)
        return summary_dict, hit_aligned_reads

    def placement(self, summary_dict):
        ## This is the placement pipeline in GraftM, in aligned reads are
        ## placed into phylogenetic trees, and the results interpreted.
        ## If reverse reads are used, this is where the comparisons are made
        ## between placements, for the summary tables to be build in the
        ## next stage.
        # Concatenate alignment files, place in tree, split output guppy
        # and .jplace file for the output
        summary_dict = self.p.place(summary_dict,
                                    self.gmf,
                                    self.args)
        # Summary steps.
        start           = timeit.default_timer()
        otu_tables      = [GraftMFiles(base, self.args.output_directory, False).summary_table_output_path(base) for base in summary_dict['base_list']]
        placements_list = []
        for base in summary_dict['base_list']:
            # First assign the hash that contains all of the trusted placements
            # to a variable to it can be passed to otu_builder, to be written
            # to a file. :)
            if summary_dict['reverse_pipe']:
                placements = summary_dict[base]['comparison_hash']['trusted_placements']
                summary_dict[base]['read_length'] = (summary_dict[base]['forward']['read_length'] + summary_dict[base]['reverse']['read_length'])/2
            elif not summary_dict['reverse_pipe']:
                placements = summary_dict[base]['trusted_placements']
            else:
                raise Exception('Programming Error: Assigning placements hash')
                
            self.s.readTax(placements, GraftMFiles(base, self.args.output_directory, False).read_tax_output_path(base))
            placements_list.append(placements)

        # Generate coverage table
        #logging.info('Building coverage table for %s' % base)
        #self.s.coverage_of_hmm(self.args.aln_hmm_file,
        #                         self.gmf.summary_table_output_path(base),
        #                         self.gmf.coverage_table_path(base),
        #                         summary_dict[base]['read_length'])
        logging.info('Building summary table')
        self.s.otu_builder(placements_list,
                           otu_tables,
                           summary_dict['base_list'],
                           self.gmf.combined_summary_table_output_path())
        
        logging.info('Building summary krona plot')
        self.kb.otuTablePathListToKrona(otu_tables,
                                        self.gmf.krona_output_path())
        stop = timeit.default_timer()
        summary_dict['summary_t'] = str(int(round((stop - start), 0)) )

        # Compile basic run statistics if they are wanted
        summary_dict['stop_all'] = timeit.default_timer()
        summary_dict['all_t'] = str(int(round((summary_dict['stop_all'] - summary_dict['start_all']), 0)) )
        self.s.build_basic_statistics(summary_dict, self.gmf.basic_stats_path(), self.args.type)
        # Delete unnecessary files
        logging.info('Cleaning up')
        for base in summary_dict['base_list']:
            directions = ['forward', 'reverse']
            if summary_dict['reverse_pipe']:
                for i in range(0,2):
                    self.gmf = GraftMFiles(base, self.args.output_directory, directions[i])
                    self.hk.delete([self.gmf.for_aln_path(base),
                                    self.gmf.rev_aln_path(base),
                                    self.gmf.conv_output_rev_path(base),
                                    self.gmf.conv_output_for_path(base),
                                    self.gmf.euk_free_path(base),
                                    self.gmf.euk_contam_path(base),
                                    self.gmf.readnames_output_path(base),
                                    self.gmf.sto_output_path(base),
                                    self.gmf.orf_titles_output_path(base),
                                    self.gmf.orf_hmmsearch_output_path(base),
                                    self.gmf.hmmsearch_output_path(base),
                                    self.gmf.orf_output_path(base),
                                    self.gmf.comb_aln_fa(),
                                    self.gmf.output_for_path(base),
                                    self.gmf.output_rev_path(base)])
            elif not summary_dict['reverse_pipe']:
                self.gmf = GraftMFiles(base, self.args.output_directory, False)
                self.hk.delete([self.gmf.for_aln_path(base),
                                self.gmf.rev_aln_path(base),
                                self.gmf.conv_output_rev_path(base),
                                self.gmf.conv_output_for_path(base),
                                self.gmf.euk_free_path(base),
                                self.gmf.euk_contam_path(base),
                                self.gmf.readnames_output_path(base),
                                self.gmf.sto_output_path(base),
                                self.gmf.orf_titles_output_path(base),
                                self.gmf.hmmsearch_output_path(base),
                                self.gmf.orf_hmmsearch_output_path(base),
                                self.gmf.orf_output_path(base),
                                self.gmf.comb_aln_fa(),                                    
                                self.gmf.output_for_path(base),
                                self.gmf.output_rev_path(base)])

        logging.info('Done, thanks for using graftM!\n')

    def graft(self):
        # The Graft pipeline:
        # Searches for reads using hmmer, and places them in phylogenetic
        # trees to derive a community structure.
        print '''
                                GRAFT
        
                       Joel Boyd, Ben Woodcroft
                       
                                                         __/__
                                                  ______|
          _- - _                         ________|      |_____/
           - -            -             |        |____/_
           - _     >>>>  -   >>>>   ____|          
          - _-  -         -             |      ______
             - _                        |_____|
           -                                  |______
            '''
        readstoplace=False # An extra check to make sure there are reads to place with
        # Set up a dictionary that will record stats as the pipeline is running
        summary_table = {'euks_checked'      : self.args.euk_check,
                         'resolve_placements': self.args.resolve_placements,
                         'base_list'         : [],
                         'seqs_list'         : [],
                         'start_all'         : timeit.default_timer(),
                         'reverse_pipe'      : False}
        
        # Set the output directory if not specified and create that directory
        logging.debug('Creating working directory: %s' % self.args.output_directory)
        self.hk.make_working_directory(self.args.output_directory,
                                       self.args.force)
        
        
        
        
        # For each pair (or single file passed to GraftM)
        logging.debug('Working with %i file(s)' % len(self.sequence_pair_list))
        for pair in self.sequence_pair_list:
            # Set the basename, and make an entry to the summary table.
            base = os.path.basename(pair[0]).split('.')[0]
            # Set reverse pipe if more than one pair
            if hasattr(self.args, 'reverse'):
                summary_table['reverse_pipe'] = True
                summary_table[base]           = {'reverse':{}, 
                                                 'forward':{}}
                pair_direction                = ['forward', 'reverse']
            else:
                summary_table[base] = {}
            
            # Set pipeline and evalue by checking HMM format
            hmm_type, hmm_tc = self.hk.setpipe(self.args.aln_hmm_file)
            logging.debug("HMM type: %s Trusted Cutoff: %s" % (hmm_type, hmm_tc))
            setattr(self.args, 'type', hmm_type)
            if hmm_tc:
                setattr(self.args, 'eval', '--cut_tc')
                
            # Guess the sequence file type, if not already specified to GraftM
            if hasattr(self.args, 'input_sequence_type'): 
                pass
            else:
                setattr(self.args, 'input_sequence_type',
                        self.hk.guess_sequence_type(pair[0],
                                                    self.input_file_format))
            logging.debug("Set sequence type of %s to %s" %(pair[0], self.args.input_sequence_type))
            # Make the working base directory
            self.hk.make_working_directory(os.path.join(self.args.output_directory,
                                                        base),
                                           self.args.force)

            # tell the user which file/s is being processed
            logging.info("Working on %s" % base)

            # for each of the paired end read files
            for read_file in pair:
                if not os.path.isfile(read_file): # Check file exists
                    logging.info('%s does not exist! Skipping this file..' % read_file)
                    continue
                # Set the output file_name
                if summary_table['reverse_pipe']:
                    direction = pair_direction.pop(0)
                    logging.info("Working on %s reads" % direction)
                    self.gmf = GraftMFiles(base,
                                           self.args.output_directory,
                                           direction)
                    self.hk.make_working_directory(os.path.join(self.args.output_directory,
                                                                base,
                                                                direction),
                                                   self.args.force)
                elif not summary_table['reverse_pipe']:
                    direction = False
                    self.gmf = GraftMFiles(base,
                                           self.args.output_directory,
                                           direction)
                else:
                    raise Exception('Programming Error')
                
                if self.args.type == 'P':
                    logging.debug("Running protein pipeline")
                    summary_table, hit_aligned_reads = self.protein_pipeline(base,
                                                                            summary_table,
                                                                            read_file,
                                                                            direction)
                # Or the DNA pipeline
                elif self.args.type == 'D':
                    logging.debug("Running nucleotide pipeline")
                    self.hk.set_euk_hmm(self.args)
                    summary_table, hit_aligned_reads = self.dna_pipeline(base,
                                                                        summary_table,
                                                                        read_file,
                                                                        direction)
                if not hit_aligned_reads:
                    continue
                else:
                    readstoplace=True
                    
                # Add the run stats and the completed run to the summary table
                summary_table['seqs_list'].append(hit_aligned_reads)
                if base not in summary_table['base_list']:
                    summary_table['base_list'].append(base)

        # Leave the pipeline if search only was specified
        if self.args.search_and_align_only:
            logging.info('Stopping before placement\n')
            exit(0)
        elif not readstoplace:
            logging.error('No hits in any of the provided files. Cannot continue with no reads to place.\n')
            exit(0)
        # Tell the user we're on to placing the sequences into the tree.
        self.gmf = GraftMFiles('',
                               self.args.output_directory,
                               False)
        logging.info("Placing reads into phylogenetic tree")
        self.placement(summary_table)


    def main(self):

        if self.args.subparser_name == 'graft':
            self.graft()

        elif self.args.subparser_name == 'assemble':
            print '''
                           ASSEMBLE

                   Joel Boyd, Ben Woodcroft

          _- - _               ___            __/
           -                  /___\____      /\/
           - _     >>>>   ___/       \_\     \/
          - _-           /_/            \    /
             - _        /                \__/
                       /
'''
            self.tg.main(self.args)

        elif self.args.subparser_name == 'extract':
            print '''
                           EXTRACT

                   Joel Boyd, Ben Woodcroft
                                  _
                         __/__/    |        >a
                  ______|          |->>>>   --------
         ________|      |_____/   _|        >b
        |        |____/_                    -------
    ____|                                   >c
        |      ______                       ----------
        |_____|
              |______
'''

            if self.args.seq:
                self.e.extract(self.args)

        elif self.args.subparser_name == 'create':
            print '''
                            CREATE

                   Joel Boyd, Ben Woodcroft
                     
                                                    /                
              >a                                   /
              -------------                       /            
              >b                        |        |
              --------          >>>     |  GPKG  |
              >c                        |________|
              ----------     
'''

            self.hk.checkCreatePrerequisites()
            Create().main(self.args.hmm, self.args.alignment, self.args.sequences, self.args.taxonomy, self.args.tree, self.args.log, self.args.output)
    



import csv
import subprocess
import tempfile

from graftm.housekeeping import HouseKeeping

class OtuTable:
    def __init__(self, sample_name):
        self.sample_name = sample_name
        self.sampleCounts = {} #taxonomy to count info

class KronaBuilder:
    def __init__(self):
        self.hk = HouseKeeping()

    def otuTablePathListToKrona(self, otuTablePaths, outputName):
        otuTables = []
        for path in otuTablePaths:
            for table in self.parseOtuTable(path):
                otuTables.append(table)
        self.runKrona(otuTables, outputName)

    def parseOtuTable(self, otuTablePath):
        data = csv.reader(open(otuTablePath), delimiter="\t")

        # Parse headers (sample names)
        fields = data.next()
        if len(fields) < 3: raise "Badly formed OTU table %s" % otuTablePath
        tables = []
        for i in range(len(fields)-2):
            table = OtuTable(fields[i+1])
            tables.append(table)

        # Parse the data in
        taxonomyColumn = len(fields)-1
        for row in data:
            for i in range(len(fields)-2):
                taxonomy = row[taxonomyColumn]
                tables[i].sampleCounts[taxonomy] = row[i+1]

        return tables

    def runKrona(self, otuTables, outputName):
        # write out the tables to files
        tempfiles = []
        tempfile_paths = []
        for table in otuTables:
            tmps = tempfile.mkstemp('','CommunityMkrona')
            tmp = tmps[1]
            out = open(tmp,'w')
            tempfiles.append(out)
            tempfile_paths.append(tmp)
            for taxonomy, count in table.sampleCounts.iteritems():
                tax = "\t".join(taxonomy.split(';'))
                out.write("%s\t%s\n" % (count,tax))
            out.close()

        cmd = ["ktImportText",'-o',outputName]
        for i, tmp in enumerate(tempfile_paths):
            cmd.append(','.join([tmp,otuTables[i].sample_name]))

        # run the actual krona
        subprocess.check_call(' '.join(cmd) + ' 1>/dev/null ', shell=True)

        # close tempfiles
        for t in tempfiles:
            t.close()






#!/usr/bin/env python

import subprocess
import os
import json
import shutil
import tempfile
import logging

import graftm.getaxnseq 
from Bio import SeqIO
from graftm.hmmer import Hmmer


class Create:
    
    def __init__(self): 
        self.h=Hmmer(None, None)
        self.the_trash=[]
    
    def buildHmm(self, alignment, base): 
        counter=0
        if os.path.isfile(base + ".hmm"):
            counter=0
            while os.path.isfile(base + ".hmm"):
                base=base+'_%s' % (str(counter))
                counter+=1
            hmm = base + ".hmm"
        else:   
            hmm = base + ".hmm" # Set a name for a hmm
        cmd = "hmmbuild %s %s >/dev/null" % (hmm, alignment) # Build the command to build the hmm
        logging.debug("Calling command: %s" % (cmd))
        subprocess.check_call(cmd, shell=True) # Call the command
        return hmm
    
    def pipeType(self, hmm):
        logging.debug("Setting pipeline type")
        type=[x.split() for x in open(hmm).readlines() if x.startswith('ALPH') or x.startswith('LENG')]
        for item in type:
            if item[0]=='ALPH':
                if item[1]=='amino':
                    ptype='aa'
                elif item[1]=='DNA' or 'RNA':
                    ptype='na'
                else:
                    raise Exception("Unfamiliar HMM type: %s" % (item[1]))
            elif item[0]=='LENG':
                leng=item[1]
            else:
                raise Exception("Programming Error: Misread HMM file")
        logging.debug("Set pipeline type as: %s " % ptype)
        logging.debug("Found alignment type as: %s" % leng)
        return ptype, leng
    
    def checkAlnLength(self, alignment):
        return len(list(SeqIO.parse(open(alignment, 'r'), 'fasta'))[0].seq)
        
    def alignSequences(self, hmm, sequences, base): 
        stockholm_alignment = base +".aln.sto" # Set an output path for the alignment
        fasta_alignment = base+".insertions.aln.fa" # Set an output path for the alignment
        corrected_fasta_alignment = base+".aln.fa" # Set an output path for the alignment
        cmd = "hmmalign --trim -o %s %s %s" % (stockholm_alignment, hmm, sequences) # Build the command to align the sequences
        logging.debug("Calling command %s" % (cmd))
        subprocess.check_call(cmd, shell=True) # Call the command
        cmd = "seqmagick convert %s %s" % (stockholm_alignment, fasta_alignment)
        logging.debug("Calling command %s" % (cmd))
        subprocess.check_call(cmd, shell=True) # Call the command
        logging.debug("Correcting alignment")
        self.h.alignment_correcter([fasta_alignment], corrected_fasta_alignment)
        self.the_trash += [stockholm_alignment, corrected_fasta_alignment, fasta_alignment]
        logging.debug("Wrote corrected alignments to: %s" % (corrected_fasta_alignment))
        return corrected_fasta_alignment
    
    def buildTree(self, alignment, base, ptype): 
        log_file = base + ".tre.log"
        tre_file = base + ".tre"
        if ptype == 'na': # If it's a nucleotide sequence
            cmd = "FastTreeMP -quiet -gtr -nt -log %s %s > %s 2>/dev/null" % (log_file, alignment, tre_file)
            logging.debug("Calling command: %s" % (cmd))
            subprocess.check_call(cmd, shell=True) # Call the command
        else: # Or if its an amino acid sequence
            cmd = "FastTreeMP -quiet -log %s %s > %s 2>/dev/null" % (log_file, alignment, tre_file)
            logging.debug("Calling command: %s" % (cmd))
            subprocess.check_call(cmd, shell=True) # Call the command
            
        self.the_trash += [log_file, tre_file]
        return log_file, tre_file

    def callTaxitCreate(self, base, aln_file, tre, tre_log, tax, seq, prefix):
        if prefix:
            refpkg = prefix + ".refpkg"
        else:
            refpkg = base + ".refpkg"
        cmd = "taxit create --quiet -f %s -P %s -t %s -s %s -c -l  %s -T %s -i %s 1>/dev/null" % (aln_file, refpkg, tre, tre_log, base, tax, seq)
        logging.debug("Calling command: %s" % cmd)
        subprocess.check_call(cmd, shell=True)
        return refpkg
    
    def compile(self, base, refpkg, hmm, contents, prefix): 
        if prefix:
            gpkg = prefix + ".gpkg"
        else:
            gpkg = base + ".gpkg"
        if os.path.isdir(gpkg): 
            raise Exception("Detected gpkg with name %s" % (gpkg))
        os.mkdir(gpkg)
        os.rename(refpkg, os.path.join(gpkg, refpkg))
        shutil.copyfile(hmm, os.path.join(gpkg, os.path.basename(hmm)))
        json.dump(contents, open(os.path.join(gpkg, 'CONTENTS.json'), 'w'))
        return

    def cleanup(self, the_trashcan):
        for every_piece_of_junk in the_trashcan:
            if os.path.isdir(every_piece_of_junk):
                shutil.rmtree(every_piece_of_junk)
            else:
                os.remove(every_piece_of_junk)
        
    def main(self, hmm, alignment, sequences, taxonomy, tree, log, prefix): 
        if sequences:
            base=os.path.basename(sequences).split('.')[0]
        else:
            base=os.path.basename(alignment).split('.')[0]
        logging.info("Building gpkg for %s" % base)
        # Initially, build the HMM if one is not provided.
        if not tree and not log:
            logging.debug("No tree or log provided")
            if hmm and alignment:
                logging.debug("Found HMM and alignment")
                ptype,leng=self.pipeType(hmm)
                if str(self.checkAlnLength(alignment)) != str(leng):
                    logging.info("Alignment length does not match the HMM length, building a new HMM")
                    hmm=self.buildHmm(alignment, base)
                    logging.info("Aligning to HMM built from alignment")
                    output_alignment = self.alignSequences(hmm, sequences, base)
                else:
                    logging.info("Alignment length matches the HMM length, no need to align")
                    output_alignment=alignment
            elif hmm and not alignment:
                logging.debug("Found HMM but not alignment")
                logging.info("Using provided HMM to align sequences")
                output_alignment = self.alignSequences(hmm, sequences, base)
            elif alignment and not hmm:
                logging.debug("Found alignment but not HMM")
                logging.info("Building HMM from alignment")
                hmm=self.buildHmm(alignment, base)
                logging.info("Aligning to HMM built from alignment")
                output_alignment = self.alignSequences(hmm, sequences, base)
            ptype,leng=self.pipeType(hmm)
            # Build the tree
            logging.info("Building tree")
            log_file, tre_file = self.buildTree(output_alignment, base, ptype)
        else:
            if not tree:
                raise Exception("No Tree provided for log file")
            if not log:
                raise Exception("No log provided for tree")
            logging.debug("Found tree and log file for tree")
            try:
                output_alignment=alignment
                tre_file=tree
                log_file=log
            except:
                raise Exception("No alignment file provided")

        # Create tax and seqinfo .csv files
        logging.info("Building seqinfo and taxonomy file")
        seq, tax = graftm.getaxnseq.main(base, taxonomy)
        self.the_trash += [seq, tax]
        
        # Create the reference package
        logging.info("Creating reference package")
        refpkg = self.callTaxitCreate(base, output_alignment, tre_file, log_file, tax, seq, prefix)

        # Compile the gpkg
        logging.info("Compiling gpkg")
        contents = {"aln_hmm": hmm,
                    "search_hmm": [hmm],
                    "rfpkg": refpkg,
                    "TC":False}
        self.compile(base, refpkg, hmm, contents, prefix)

        logging.info("Cleaning up")
        self.cleanup(self.the_trash)
        
        logging.info("Finished\n")
        return



from skbio import TreeNode

class Rerooter:
    def __init_(self): pass

    def importer(self, tree_path):
        t = TreeNode.from_newick(open(tree_path, 'r'))
        return t

    def node_list(self, t, nodes_path):
        id_list = []
        misses = []
        for i in open(nodes_path).readlines():
            try:
                id_list.append(t.find(i.rstrip()))
            except:
                misses.append(i.rstrip())

        if len(id_list) < 1:
            print 'No nodes to continue with. Exiting'
            exit()
        if len(misses) > 0:
            print 'The following nodes were not found in the tree, continuing with %s nodes:\n%s' % (len(id_list), str(' '.join(misses)))

        return id_list

    def lca(self, t, n):
        lca = t.lowest_common_ancestor(n)
        return lca

    def reroot(self, lca, t):
        root = lca.root()
        return root

    def write_tree(self, t, o):
        t.write(o, format='newick')

    def main(self, i, o, n):

        #load tree
        tree = self.importer(i)

        # Get nodes
        nodes = self.node_list(tree, n)

        # Find last common ancestor
        lca = self.lca(tree, nodes)

        # Reroot dat tree
        rerooted_tree = self.reroot(lca, tree)

        # Write to file
        self.write_tree(rerooted_tree , o)



import subprocess
import tempfile
import os
import shutil
import logging

from Bio import SeqIO

class TaxoGroup:

    def __init__(self): pass

    def main(self, args):
        ## Split the lists of graftM runs

        # Define a list to contain the paths to all of the guppy files that will be used
        guppy_list= []

        # Define a list to contan the paths to all of the sequence files that will be used
        f_sequence_list = []

        # Split the comma separated input files
        file_list = args.graft_run.split(',')

        #
        for run in file_list:
            guppy_list.append(run + '/' + run + '_placements.guppy')
            f_sequence_list.append(run + '/' + run + '_hits.fa')

        ## Creates a list of sequences and their taxonomic classification from all the guppy files
        logging.info("Preparing reads")
        split_guppys = TaxoGroup().guppy_splitter(guppy_list)

        ## Grabs the sequences from the forward reads, and grabs the corresponding reverse reads as well.
        grouped_sequences = TaxoGroup().rank_grouper(split_guppys, f_sequence_list)

        logging.info("Assembling")
        TaxoGroup().assembler(grouped_sequences, 'graftM_assemble', args.assembly_type, args)

        exit(1)

    def guppy_splitter(self, guppy_file, cutoff):
        ## Split a gupyp file, and return the placement information in hash
        ## format, and also the placement confidence at each rank

        # Send the message that the guppy file is being split
        logging.info("Parsing the guppy file(s)")
        read_tax_dict = {} # Define an empty output hash
        
        # Split each line in the guppy, check that it meets the cutoff, and
        # Return a hash with the placements, and rank placements
        for lst in [x.replace('Candidatus ', 'Candidatus_').split() for x in open(guppy_file,'r')]: # for each entry in the guppy file
            if lst[0] != 'name': # ignore the title entry of the guppy
                lst[0] = '_'.join(lst[0].split('_')[:-1]) # Remove the unique identifier of this read name
                if lst[1] == lst[2] and float(lst[len(lst)-2]) > float(cutoff): # If the placements match, and the placement confidence meets the cutoff
                    if lst[0] not in read_tax_dict: # If the read name is not already in the list
                        read_tax_dict[lst[0]] = {}
                        read_tax_dict[lst[0]]['placement'] = [lst[3]] # Add a placement entry for this read
                        read_tax_dict[lst[0]]['confidence'] = [lst[len(lst)-2]] # Add a confidence entry for this read
                    else: # else, if there is already an entry for this read
                        if len(read_tax_dict[lst[0]]['placement']) < 7: # Ensure that the rank of placement isn't already full
                            read_tax_dict[lst[0]]['placement'].append(lst[3]) # And append the rank to the list.
                            read_tax_dict[lst[0]]['confidence'].append(lst[len(lst)-2]) # And append the confidence to the list
                        else:
                            # Go out on a limb here, and assume that the full placement has been given. Sometimes placements are duplicated for the same read in the guppy file.
                            # raise Exception('Programming Error: Guppy file not formatted correctly: %s' % (guppy_file)) # If its full, and there are too many placements for a given read, fail.
                            pass
        # Return the hash.
        return read_tax_dict

    def rank_grouper(self, split_guppy, f_sequences_list):
        '''Gathers the sequences that correspond to each id, and taxonomic rank within the split guppy'''
        n_misses = 0
        tax_list = [{}, {}, {}, {}, {}, {}]

        f_sequences = {}


        for f in f_sequences_list:
            records = SeqIO.to_dict(SeqIO.parse(open(f, 'r'), "fasta"))
            f_sequences.update(records)
            logging.info( "%s reads found in %s" % (len( records.keys() ), os.path.basename(f) ) )

        logging.info("Using %s forward reads in total." % (len(f_sequences.keys())))
        logging.info("Sorting 16S forward sequences into taxonomic groups")

        for read_name, classification in split_guppy.iteritems():
            taxonomic_group = classification[len(classification) - 1]

            tax_dict = tax_list[len(classification) - 2]

            if taxonomic_group not in tax_dict:
                tax_dict[taxonomic_group] = []
            tax_dict[taxonomic_group].append(f_sequences[read_name[:-2]])

        return tax_list

    def assembler(self, tax_rank_list, output_directory, assembly_type, args):

        rank = ['K','P','C','O','F','G']
        p_rank= ['Kingdom','Phylum','Class','Order','Family','Genus']
        cmd = 'mkdir %s 2>/dev/null' % output_directory

        try:
            subprocess.check_call(cmd, shell = True)

        except:
            logging.info("A folder with the name '%s' already exists... This is awkward... \n" % output_directory)
            exit(1)


        for idx, taxonomic_rank in enumerate(tax_rank_list):
            r = p_rank.pop(0)
            if r == 'Genus':
                logging.info("Using %s to assemble reads classified at the %s level." % (assembly_type, r))
            else:
                logging.info("Using %s to assemble reads classified at the %s level. -WARNING- These assemblies are likely to be rubbish." % (assembly_type, r))

            for tax_name, sequences in taxonomic_rank.iteritems():
                number_of_sequences = len(list(sequences))
                if number_of_sequences > 1:
                    output_assembly_file_path = os.path.join(output_directory, '%s_%s_%s_assembly.fa' % (rank[idx], tax_name, str(number_of_sequences)))
                    output_reads_file_path = os.path.join(output_directory, '%s_%s_%s_reads.fa' % (rank[idx], tax_name, str(number_of_sequences)))
                    velvet_assembly_path = os.path.join(output_directory, '%s_%s_velvet_assembly' % (rank[idx],tax_name))

                    if assembly_type == "phrap":

                        with tempfile.NamedTemporaryFile() as tmp:

                            SeqIO.write(sequences, output_reads_file_path, "fasta")
                            cmd = 'phrap -minscore 40 -minmatch 45 -revise_greedy -forcelevel 0 -repeat_stringency 0.99 %s >/dev/null' % (output_reads_file_path)

                            subprocess.check_call(cmd, shell = True)

                            cmd = 'mv %s.contigs %s' % (output_reads_file_path, output_assembly_file_path)
                            subprocess.check_call(cmd, shell = True)

                    elif assembly_type == "finishm":

                        # Write out the sequences to this file
                        SeqIO.write(sequences, output_reads_file_path, "fasta")


                        # Run finishm visualise on the clustered sequences
                        cmd = 'finishm visualise --quiet --assembly-coverage-cutoff 3.5 --assembly-kmer %s --velvet-directory %s --fasta %s --assembly-svg %s' % (args.kmer, velvet_assembly_path, output_reads_file_path, output_assembly_file_path.replace('.fa', '.svg'))
                        subprocess.check_call(cmd, shell = True)

                        # Move the output picture to the output directory
                        cmd = 'mv *.svg %s' % (output_directory)
                        subprocess.check_call(cmd, shell = True)

                        # Move the clustered sequence files to the output directory
                        cmd = 'mv %s %s' % (output_sequence_file_path.replace('_assembly.fa', ''), output_directory)
                        subprocess.check_call(cmd, shell = True)

                        # Assemble the reads, as best as finishm can
                        cmd = 'finishm assemble --quiet --already-assembled-velvet-directory %s/%s --output-contigs %s --output-pathspec --bubbly --no-progressbar --max-bubble-size 999999999 --post-assembly-coverage-cutoff 10.5' % (output_directory, output_sequence_file_path, output_sequence_file_path.replace('_assembly.fa', '.fa'), )
                        subprocess.check_call(cmd, shell = True)

                        # Move the resulting assembly to the output directory
                        cmd = 'mv %s %s' % (output_sequence_file_path.replace('_assembly.fa', ''), output_directory)
                        subprocess.check_call(cmd, shell = True)

                    elif assembly_type == "velvet":

                        # Write out the sequences to this file
                        SeqIO.write(sequences, output_reads_file_path, "fasta")

                        # Run a velvet assembly
                        cmd = 'velveth %s %s -fasta -short %s 1>/dev/null' % (velvet_assembly_path, args.kmer, output_reads_file_path)
                        subprocess.check_call(cmd, shell = True)
                        cmd = 'velvetg %s -cov_cutoff 3.5 1>/dev/null' % (velvet_assembly_path)
                        subprocess.check_call(cmd, shell = True)



                        if args.finish:

                            if len([x for x in open(os.path.join(velvet_assembly_path, 'contigs.fa'), 'r').readlines() if x.startswith('>')]) > 1:
                                # Finish off fragments
                                cmd = 'phrap -minscore 40 -minmatch 45 -revise_greedy -forcelevel 0 -repeat_stringency 0.99 %s > /dev/null 2>&1' % (os.path.join(velvet_assembly_path, 'contigs.fa'))
                                subprocess.check_call(cmd, shell = True)

                                # Move the resulting assembly to the output directory
                                shutil.move(os.path.join(velvet_assembly_path, 'contigs.fa.contigs'),  output_assembly_file_path)

                            elif len([x for x in open(os.path.join(velvet_assembly_path, 'contigs.fa'), 'r').readlines() if x.startswith('>')]) == 1:
                                # Move the resulting assembly to the output directory
                                shutil.move(os.path.join(velvet_assembly_path, 'contigs.fa'),  output_assembly_file_path)

                            else:
                                pass
                        else:

                            # Move the resulting assembly to the output directory
                            shutil.move(os.path.join(velvet_assembly_path, 'contigs.fa'),  output_assembly_file_path)

        logging.info("Finished assembling\n")




import random
import os
import shutil
import subprocess
import json
import tempfile
import logging
from timeit import itertools

# Constants - don't change them evar.
FORMAT_FASTA = 'FORMAT_FASTA'
FORMAT_FASTQ_GZ = 'FORMAT_FASTQ_GZ'

class UninstalledProgramError(Exception):
    pass

class HouseKeeping:
    ### Functions for setting up the graftM pipeline to run correctly, and 
    ### general housekeeping things like adding and removing directories and 
    ### files.
    
    def __init__(self): pass

    def contents(self, path):
        contents = os.path.join(path, 'CONTENTS.json')
        if os.path.isfile(contents):
            return json.load(open(contents, 'r'))
        else:
            return None
        
    def guess_sequence_input_file_format(self, sequence_file_path):
        '''Given a sequence file, guess the format and return. Raise an 
        exception if it cannot be guessed'''
        if sequence_file_path.endswith(('.fa', '.faa', '.fna', '.fasta')):  # Check the file type
            return FORMAT_FASTA
        elif sequence_file_path.endswith(('.fq.gz', '.fastq.gz', 'fasta.gz', 'fa.gz', 'faa.gz', 'fna.gz')):
            return FORMAT_FASTQ_GZ
        else:
            raise Exception("Unable to guess file format of sequence file: %s" % sequence_file_path)
        
    def guess_sequence_type(self, input_file, file_format):
        '''Guess the type of input sequence provided to graftM (i.e. nucleotide
        or amino acid) and return'''
        
        # Define expected residues for each sequence type
        aas = set(['P','V','L','I','M','F','Y','W','H','K','R','Q','N','E','D','S'])
        nas = set(['A', 'T', 'G', 'C', 'N', 'U'])
        
        # If its Gzipped and fastq make a small sample of the sequence to be 
        # read
        if file_format == FORMAT_FASTQ_GZ:
            return 'nucleotide'
        else:
            with open(input_file) as in_file:
                head = [next(in_file).rstrip() for x in xrange(2)]
                for nucl in set(head[1]):
                    if nucl not in nas and nucl in aas:
                        return 'protein'
                    elif nucl not in nas and nucl not in aas:
                        raise Exception(logging.error('Encountered unexpected character when attempting to guess sequence type: %s' % (nucl)))
                    else:
                        continue
                return 'nucleotide'

    def set_euk_hmm(self, args):
        'Set the hmm used by graftM to cross check for euks.'
        if hasattr(args, 'euk_hmm_file'):
            pass
        elif not hasattr(args, 'euk_hmm_file'):
            setattr(args, 'euk_hmm_file', os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), 'constants', '18S.hmm'))
        else:
            raise Exception('Programming Error: setting the euk HMM')    

    
    def setpipe(self, hmm):
        ## Read in the hmm file and return the evalue trusted cutoff and the
        ## type of HMM
        hmm_tc = None
        hmm_read = [x.rstrip() for x in open(hmm, 'r').readlines() if x.startswith('ALPH') or x.startswith('TC')]
        if len(hmm_read) > 2:
            raise Exception('Possibly Programming Error: Misread HMM library.')
        for item in hmm_read:
            if item.startswith('ALPH'):
                if item.split()[1] == 'DNA' or item.split()[1] == 'RNA':
                    hmm_type = 'D'
                elif item.split()[1] == 'amino':
                    hmm_type = 'P'
            elif item.startswith('TC'):
                hmm_tc = item.split()[1]
            else:
                raise Exception('Possibly Programming Error: Misread HMM library.')
        return hmm_type, hmm_tc
        
    def delete(self, delete_list):
        for item in delete_list:
            try:
                os.remove(item)
            except:
                pass

    def make_working_directory(self, directory_path, force):
        if force:
            shutil.rmtree(directory_path, ignore_errors=True)
            os.mkdir(directory_path)
        else:
            try:
                os.mkdir(directory_path)
            except:
                logging.error('Directory %s already exists. Exiting to prevent over-writing' % directory_path)
                raise Exception('Directory %s already exists. Exiting to prevent over-writing'% directory_path)
    
    def parameter_checks(self, args):
        ## Check that the necessary files are in place
        if args.subparser_name == 'graft':
            # Check that the placement cutoff is between 0.5 and 1
            if float(args.placements_cutoff) < float(0.5) or float(args.placements_cutoff) > float(1.0):
                logging.info('Please specify a confidence level (-d) between 0.5 and 1.0! Found: %s' % args.placements_cutoff)
                exit(1)

            # Set string for hmmsearch evalue
            args.eval = '-E %s' % args.eval
            
            self._check_file_existence(args.forward)
            if hasattr(args, 'reverse'):
                self._check_file_existence(args.reverse)
                    
            # Determine the File format based on the suffix
            input_file_format = self.guess_sequence_input_file_format(args.forward[0])
            logging.debug("Detected file format %s" % input_file_format)
            sequence_file_list = []
            if hasattr(args, 'reverse'):
                if len(args.forward) != len(args.reverse):
                    logging.error('Confusing input. There appears to be different numbers of forward and reverse files specified')
                for i, forward_file in enumerate(args.forward):
                    sequence_file_list.append([forward_file, args.reverse[i]])
            else:
                sequence_file_list = [[f] for f in args.forward]
                
            return sequence_file_list, input_file_format
        else:
            return
        
    def _check_file_existence(self, files):
        '''Iterate through files and exit(1) if any do not pass
        os.path.isfile'''
        for f in files:
            if not os.path.isfile(f):
                logging.error("The file '%s' does not appear to exist, stopping" % f)
                exit(1)
            
    
    def checkCreatePrerequisites(self):
        uninstalled_programs = []
        prerequisites = {'taxit': 'https://github.com/fhcrc/taxtastic',
                         'FastTree': 'http://www.microbesonline.org/fasttree/',
                         'seqmagick': 'https://github.com/fhcrc/seqmagick'}
        for program in prerequisites.keys():
            if self.which(program):
                pass
            else:
                uninstalled_programs.append(program)
        if uninstalled_programs:
            msg = "The following programs must be installed to run GraftM create\n"
            logging.info(msg)
            for program in uninstalled_programs:
                l = '\t%s\t%s' % (program, prerequisites[program])
                print l
                msg += l+"\n"
            exit(0)
            
    def set_attributes(self, args):
        # Check the presence of all prerequisite programs needed for GraftM
        uninstalled_programs = []
        prerequisites = {'orfm': 'https://github.com/wwood/OrfM',
                        'nhmmer': 'http://hmmer.janelia.org/',
                        'hmmsearch': 'http://hmmer.janelia.org/',
                        'fxtract': 'https://github.com/ctSkennerton/fxtract',
                        'pplacer': 'http://matsen.fhcrc.org/pplacer/',
                        'seqmagick': 'https://github.com/fhcrc/seqmagick',
                        'ktImportText': 'http://sourceforge.net/p/krona/home/krona/'}
        for program in prerequisites.keys():
            if self.which(program):
                pass
            else:
                uninstalled_programs.append(program)
        if uninstalled_programs:
            msg = "The following programs must be installed to run GraftM\n"
            logging.info(msg)
            for program in uninstalled_programs:
                l = '\t%s\t%s' % (program, prerequisites[program])
                print l
                msg += l+"\n"
            raise UninstalledProgramError(msg)
        else:
            pass
        
        # Read graftM package and assign HMM and refpkg file
        
        if hasattr(args, 'graftm_package'):
            if not os.path.isdir(args.graftm_package):
                raise Exception("%s does not exist. Are you sure you provided the correct path?" % args.graftm_package)
            if self.contents(args.graftm_package) is None:
                raise Exception('Misformatted GraftM package')
            elif self.contents(args.graftm_package) is not None:
                c = self.contents(args.graftm_package)
                if hasattr(args, 'search_hmm_files'): # If a hmm is specified, overwrite the one graftM package
                    setattr(args, 'aln_hmm_file', os.path.join(args.graftm_package, c['aln_hmm']))
                    setattr(args, 'reference_package', os.path.join(args.graftm_package, c['rfpkg']))
                else:
                    setattr(args, 'search_hmm_files', [])
                    for hmm in c['search_hmm']:
                        args.search_hmm_files.append(os.path.join(args.graftm_package, hmm))
                    setattr(args, 'aln_hmm_file', os.path.join(args.graftm_package, c['aln_hmm']))
                    setattr(args, 'reference_package', os.path.join(args.graftm_package, c['rfpkg']))
        elif hasattr(args, 'search_hmm_files'):
            if not hasattr(args, 'aln_hmm_file'):
                if len(args.search_hmm_files) == 1:
                    setattr(args, 'aln_hmm_file', args.search_hmm_files[0])
                else:
                    raise Exception("Multiple search HMMs specified, but aln_hmm_file not specified")
        elif hasattr(args, 'search_hmm_list_file'):
            setattr(args, 'search_hmm_files', [x.rstrip() for x in open(args.search_hmm_list_file).readlines()])

            if not hasattr(args, 'aln_hmm_file'):
                raise Exception("Multiple search HMMs specified, but aln_hmm_file not specified")
        else:
            raise Exception('No refpkg or HMM specified: Do not know what to search with.')
        return
    
    def which(self, program):
        '''Credits to BamM and http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python'''
        def is_exe(fpath):
            return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                path = path.strip('"')
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file
        return None




import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

exec(open('graftm/version.py').read()) # loads __version__

setup(name='graftm',
      version=__version__,
      author='Joel Boyd, Ben Woodcroft',
      description='GraftM is a pipeline used for identifying and classifying marker gene reads from metagenomic datasets',
      long_description=readme,
      license='see LICENSE.txt',
      keywords="",
      packages=find_packages(exclude='docs'),
      install_requires=('biopython ==1.64',
                        'seqmagick ==0.5.0',
                        'scikit-bio ==0.2.2'),
      url='http://geronimp.github.com/graftM',
      scripts=['bin/graftM']
)
